# This file should have no `from __future__ import annotations` to improve coverage of some cases
from typing import ClassVar
from unittest import TestCase

import typing

from stereotype import Model


class TestRawAnnotations(TestCase):
    def test_class_var(self):
        class HasClassVar(Model):
            normal_field: int = 1
            # ClassVar attributes shouldn't become fields
            class_var_1: ClassVar[int] = 2
            class_var_2: "ClassVar[float]" = 3.0
            class_var_3: "typing.ClassVar[typing.List[str]]" = ["4"]

            @classmethod
            def class_var_sum(cls):
                return cls.class_var_1 + cls.class_var_2

        model = HasClassVar({"class_var": "ignore me"})
        self.assertEqual(1, model.normal_field)
        self.assertEqual(2, model.class_var_1)
        self.assertEqual(3.0, model.class_var_2)
        self.assertEqual(5.0, model.class_var_sum())
        self.assertEqual(["4"], model.class_var_3)
        self.assertEqual(2, HasClassVar.class_var_1)
        self.assertEqual(3.0, HasClassVar.class_var_2)
        self.assertEqual(5.0, HasClassVar.class_var_sum())
        self.assertEqual(["4"], HasClassVar.class_var_3)
        self.assertEqual({"normal_field": 1}, model.serialize())



from collections.abc import Mapping
from copy import deepcopy
from typing import Any, Optional, Type, cast, Iterable, Sequence, Union

try:
    # Schematics 2
    from schematics.exceptions import (
        CompoundError as SchematicsValidationError,
        DataError as SchematicsConversionError,
    )
except ImportError:  # pragma: no cover
    # Schematics 1
    from schematics.exceptions import (
        ValidationError as SchematicsValidationError,
        ModelConversionError as SchematicsConversionError,
    )
from schematics.models import Model as SchematicsModel

from stereotype.fields.annotations import AnnotationResolver
from stereotype.fields.model import ModelField
from stereotype.roles import DEFAULT_ROLE, Role
from stereotype.utils import (
    Missing,
    ValidationContextType,
    PathErrorType,
    ToPrimitiveContextType,
    ConversionError,
)


class SchematicsModelField(ModelField):
    """
    Field containing a Schematics ``Model`` class, as specified in the annotation.
    Provides a way to combine stereotype and schematics models, e.g., for migrating from schematics in steps.
    Deep copy doesn't work in Schematics 2, works in Schematics 1.

    :param default: Means the field isn't required, should be None or a callable, for example the model's class
    :param hide_none: If the field's value is None, it will be hidden from serialized output
    :param primitive_name: Changes the key used to represent the field in serialized data - input or output
    :param to_primitive_name: Changes the key used to represent the field in serialized data - output only
    """

    def __init__(
        self,
        *,
        default: Any = Missing,
        hide_none: bool = False,
        primitive_name: Optional[str] = Missing,
        to_primitive_name: Optional[str] = Missing
    ):
        super().__init__(
            default=default,
            hide_none=hide_none,
            primitive_name=primitive_name,
            to_primitive_name=to_primitive_name,
        )

        self.type: Type[SchematicsModel] = cast(Type[SchematicsModel], NotImplemented)

    def init_from_annotation(self, parser: AnnotationResolver):
        if not issubclass(parser.annotation, SchematicsModel):
            raise parser.incorrect_type(self)
        self.type = parser.annotation

    def validate(
        self, value: SchematicsModel, context: ValidationContextType
    ) -> Iterable[PathErrorType]:
        try:
            value.validate()  # Cannot propagate the context to schematics
        except SchematicsValidationError as e:
            # to_primitive works for Schematics 2, messages for Schematics 1
            messages = e.to_primitive() if hasattr(e, "to_primitive") else e.messages
            yield from _iterate_validation_errors(messages)

    def copy_value(self, value: Any) -> Any:
        if value is None or value is Missing:
            return value
        return deepcopy(value)

    def to_primitive(
        self,
        value: Any,
        role: Role = DEFAULT_ROLE,
        context: ToPrimitiveContextType = None,
    ) -> Any:
        if value is None or value is Missing:
            return value
        role_str = role.name if role is not DEFAULT_ROLE else None
        value: SchematicsModel
        return value.to_primitive(role_str, context)

    def convert(self, value):
        try:
            return super().convert(value)
        except SchematicsConversionError as e:
            _raise_conversion_error(e)


def _iterate_validation_errors(messages: dict) -> Iterable[PathErrorType]:
    for key, container in messages.items():
        if isinstance(container, list):
            for error in container:
                yield (str(key),), error
        else:
            for path, error in _iterate_validation_errors(container):
                yield (str(key),) + path, error


def _raise_conversion_error(error: SchematicsConversionError) -> None:
    # Get the first error (for conversion, Schematics reports multiple, stereotype just one) and gather its path
    path = []
    first_error_field: Mapping[str, Union[Sequence[str], Mapping]] = error.args[0]
    while isinstance(first_error_field, Mapping):
        path_key, first_error_field = next(iter(first_error_field.items()))
        path.append(path_key)
    first_error_field: Sequence[str]

    # Schematics 1 -> list[str], Schematics 2 -> ConversionError, in both cases a sequence of string-like messages
    first_message = str(first_error_field[0])
    raise ConversionError.new(first_message, *path)



from __future__ import annotations

from typing import Union, List, Dict, Tuple, cast, Any, Callable


class _MissingType:
    """A special singleton type - `Missing` should be the only instance in existence."""
    def __bool__(self):
        return False

    def __repr__(self):
        return 'Missing'

    def __copy__(self):
        return Missing  # Singleton instance, shall not be copied

    def __deepcopy__(self, _):
        return Missing  # Singleton instance, shall not be copied


#: Placeholder value for required fields with missing value
#: (i.e., the field doesn't specify a default, no relation to ``Optional``).
#: Validation will always fail if this value is present, i.e. it's never present in valid models.
#:
#: In the unlikely event of needing to check for this value in code, use ``model.field is Missing``.
Missing = _MissingType()


class StereotypeError(Exception):
    pass


class ConfigurationError(StereotypeError, AssertionError):
    pass


# The validation context is an opaque value passed along to custom Field validators and validate_* methods on Models.
# The type alias serves purely as internal documentation, making types easier to read.
ValidationContextType = Any

# The to_primitive context is an opaque value passed along for use in custom Field and Model implementations.
# The type alias serves purely as internal documentation, making types easier to read.
ToPrimitiveContextType = Any

# An error message (second element) with a path of primitive field names leading to the bad data.
PathErrorType = Tuple[Tuple[str, ...], str]

# Receives the validated value as the only argument, raises ValueError if invalid.
Validator = Callable[[Any, ValidationContextType], None]


class DataError(StereotypeError):
    """Encapsulates one or multiple errors that happened during conversion or validation, mapped by field paths."""

    error_list: List[PathErrorType]

    ERRORS_DICT_SELF_KEY = '_global'

    def __init__(self, errors: List[PathErrorType]):
        self.error_list = errors
        super().__init__(self._error_string())

    def _error_string(self) -> str:
        for path, error in self.error_list:
            if not path:
                return error
            return f'{": ".join(path)}: {error}'
        assert self.error_list, 'Cannot create the exception without any errors'

    @property
    def errors(self) -> Dict[str, Union[List[str], dict]]:
        """
        Generates a potentially deeply nested dictionary with errors and their paths.

        Keys in the dictionaries are field (primitive) names.
        Values are either lists of error messages from simple fields, or recursive dictionaries for compound fields.
        """
        errors = {}
        for path, error in self.error_list:
            container = errors
            if not path:
                path = (self.ERRORS_DICT_SELF_KEY,)
            for item in path[:-1]:
                value = container.setdefault(item, {})
                if isinstance(value, list):
                    value = container[item] = cast(dict, {self.ERRORS_DICT_SELF_KEY: value})
                container = value
            array = container.setdefault(path[-1], [])
            if isinstance(array, dict):
                # The target container is not an array as it has some deeper error
                array = array.setdefault(self.ERRORS_DICT_SELF_KEY, [])
            array.append(error)
        return errors

    @classmethod
    def new(cls, message: str, *path: str):
        return cls([(path, message)])

    def wrapped(self, *path: str) -> DataError:
        raise type(self)([(path + original_path, error) for original_path, error in self.error_list])


class ConversionError(DataError, TypeError):
    """Single error created when converting raw data to a Model, caused mostly by failed type coercions."""


class ValidationError(DataError, ValueError):
    """Potentially multiple errors created when validating already converted models."""



# Parts of this file are referenced from docs by line numbers, always update docs after changing it!

from typing import List, Dict
from stereotype import Model, ListField, StrField, FloatField, ValidationError, DictField


class Book(Model):
    name: str
    authors: List[str] = ListField(min_length=1)
    reviews: Dict[str, float] = DictField(
        key_field=StrField(min_length=1),
        value_field=FloatField(min_value=0, max_value=5),
        default=dict,
    )


model = Book({'name': 'The Go Programming Language'})
model.authors = ['Alan A.A. Donovan', 'Brian W. Kernighan']
model.validate()
assert model.to_primitive() == {
    'name': 'The Go Programming Language',
    'authors': ['Alan A.A. Donovan', 'Brian W. Kernighan'],
    'reviews': {},
}


model = Book({
    'name': "The Hitchhiker's Guide to the Galaxy",
    'authors': ['Douglas Adams', None],
    'reviews': {'dolphins': 4.2, 'whales': -2},
})
assert model.reviews['dolphins'] == 4.2
try:
    model.validate()
    assert False, 'Validation fails due to the error below'  # pragma: no cover
except ValidationError as e:
    assert e.errors == {
        'authors': {'1': ['This field is required']},
        'reviews': {'whales': ['Must be between 0 and 5']},
    }



from __future__ import annotations

from copy import deepcopy, copy
from typing import Any, Optional, Callable, Iterable, TYPE_CHECKING, List, Tuple

from stereotype.fields.annotations import AnnotationResolver
from stereotype.roles import DEFAULT_ROLE, Role
from stereotype.utils import Missing, ConfigurationError, PathErrorType, ValidationContextType, Validator, \
    ToPrimitiveContextType

if TYPE_CHECKING:  # pragma: no cover
    from stereotype.model import _OutputFieldConfig, _InputFieldConfig, _ValidatedFieldConfig, _ValidatorMethod, \
        _NativeValidator, _SerializableFn


def field_method_overriden(obj, method_name) -> bool:
    return getattr(type(obj), method_name) is not getattr(Field, method_name)


class Field:
    """
    Abstract base class for other field types. Use :class:`AnyField` if type shouldn't be checked.

    :param default: Means the field isn't required, used as default directly or called if callable
    :param hide_none: If the field's value is None, it will be hidden from serialized output
    :param hide_empty: If the field's value is empty_value (varies by field type), it will be hidden
    :param primitive_name: Changes the key used to represent the field in serialized data - input or output
    :param to_primitive_name: Changes the key used to represent the field in serialized data - output only
    :param validators: Optional list of validator callbacks - they raise ``ValueError`` if the value is invalid
    """

    __slots__ = ('name', 'required', 'allow_none', 'default', 'default_factory', 'native_validate', 'validator_method',
                 'validators', 'hide_none', 'hide_empty', 'primitive_name', 'to_primitive_name', 'serializable',
                 'custom_to_primitive')
    type = NotImplemented
    type_repr: str = NotImplemented
    atomic: bool = False
    empty_value = NotImplemented

    def __init__(self, *, default: Any = Missing, hide_none: bool = False, hide_empty: bool = False,
                 primitive_name: Optional[str] = Missing, to_primitive_name: Optional[str] = Missing,
                 validators: Optional[List[Validator]] = None):
        # All NotImplemented *must* be updated later based on annotations
        self.name: str = NotImplemented
        self.allow_none: bool = False
        self.primitive_name: Optional[str] = primitive_name  # Missing falls back to field name
        self.to_primitive_name: Optional[str] = to_primitive_name  # Missing falls back to field name
        if primitive_name is not Missing and to_primitive_name is Missing:
            self.to_primitive_name = primitive_name

        validate_overridden = field_method_overriden(self, 'validate')
        self.native_validate: Optional[_NativeValidator] = self.validate if validate_overridden else None
        self.validator_method: Optional[_ValidatorMethod] = None
        self.validators: Optional[Tuple[Validator, ...]] = tuple(validators) if validators else None
        self.serializable: Optional[_SerializableFn] = None

        # Only user-specifiable options are allowed as arguments to avoid user confusion
        self.required: bool = True
        self.default: Optional[Any] = None
        self.default_factory: Optional[Callable[[], Any]] = None
        if default is not Missing:
            self.init_default(default)
        self.hide_none = hide_none
        assert not (hide_empty and self.empty_value is NotImplemented), f'{type(self)} does not support hide_empty'
        self.hide_empty = hide_empty
        self.custom_to_primitive = field_method_overriden(self, 'to_primitive')

    def init_from_annotation(self, parser: AnnotationResolver):
        """Check this Field type is appropriate for the annotation and load any nested types from it."""
        raise NotImplementedError  # pragma: no cover

    def init_name(self, name: str):
        self.name = name
        if self.primitive_name is Missing:
            self.primitive_name = name
        if self.to_primitive_name is Missing:
            self.to_primitive_name = name

    def init_default(self, default: Any):
        self.required = False
        if callable(default):
            self.default_factory = default
        else:
            self.default = default

    def check_default(self):
        """Check the default is valid input for the field. Called after `init_from_annotation` and `init_name`."""
        if self.required or self.default_factory is not None:
            pass
        elif self.default is None:
            if not self.allow_none:
                raise ConfigurationError("Cannot use None as default on a non-Optional Field")
        elif self.type is not NotImplemented and not isinstance(self.default, self.type):
            raise ConfigurationError(f'Value `{self.default}` used as field default must be of type {self.type_repr}')

    def copy_field(self):
        """
        Copies the field definition - explicit Fields must be copied, otherwise subclasses would share them.
        Any fields where plain `copy.copy` won't be enough should be manually adjusted afterwards.
        """
        copied = copy(self)
        # The native_validate slot contains a method, which would normally remain bound to the old instance
        native_validate = getattr(self, 'native_validate', None)
        if native_validate is not None:
            copied.native_validate = getattr(copied, native_validate.__func__.__name__)
        return copied

    def validation_errors(self, value: Any, context: ValidationContextType) -> Iterable[PathErrorType]:
        if value is Missing or (value is None and not self.allow_none):
            yield (), 'This field is required'
            return
        if self.native_validate is not None and value is not None:
            yield from self.native_validate(value, context)
        if self.validators is not None:
            for validator in self.validators:
                try:
                    validator(value, context)
                except ValueError as e:
                    yield (), str(e)

    def validate(self, value: Any, context: ValidationContextType) -> Iterable[PathErrorType]:
        """Method to override to add native field validation. If not overridden, native_validate will stay None."""
        yield from ()

    def _fill_missing(self):
        if self.required:
            return Missing
        if self.default_factory is None:
            return self.default
        return self.default_factory()

    def convert(self, value: Any) -> Any:
        if value is Missing:
            return self._fill_missing()
        if value is None:
            return None
        return self.type(value)

    def to_primitive(self, value: Any, role: Role = DEFAULT_ROLE, context: ToPrimitiveContextType = None) -> Any:
        return value

    def copy_value(self, value: Any) -> Any:
        return value

    def make_input_config(self) -> _InputFieldConfig:
        return self.name, self.primitive_name, self.convert, (None if self.atomic else self.copy_value)

    def make_validated_config(self) -> _ValidatedFieldConfig:
        return (self.name, self.primitive_name or self.to_primitive_name or self.name, self.allow_none,
                self.validation_errors, self.validator_method)

    def make_output_config(self) -> _OutputFieldConfig:
        # Note this isn't expected to be used if to_primitive_name is None
        return (self, self.name, self.serializable,
                self.to_primitive if self.custom_to_primitive else None, self.to_primitive_name,
                self.hide_none, self.hide_empty, self.empty_value)

    def has_validation(self) -> bool:
        return self.required or not self.allow_none or self.validator_method or self.native_validate or self.validators

    def __repr__(self):
        type_repr = f'Optional[{self.type_repr}]' if self.allow_none else self.type_repr
        return (
            f'<Field{f" {self.name}" if self.name is not NotImplemented else ""} of type {type_repr}, '
            f'{"required" if self.required else f"default=<{self.default}>"}'
            f'{f", primitive name {self.primitive_name}" if self.primitive_name != self.name else ""}'
            f'{f", to output {self.to_primitive_name}" if self.to_primitive_name != self.primitive_name else ""}>'
        )


class AnyField(Field):
    """
    Value of any type (usually annotation ``typing.Any``, but can be anything).

    :param deep_copy: If true, conversion, serialization and copying will use copy.deepcopy for this value
    :param default: Means the field isn't required, used as default directly or called if callable
    :param hide_none: If the field's value is None, it will be hidden from serialized output
    :param primitive_name: Changes the key used to represent the field in serialized data - input or output
    :param to_primitive_name: Changes the key used to represent the field in serialized data - output only
    :param validators: Optional list of validator callbacks - they raise ``ValueError`` if the value is invalid
    """

    __slots__ = Field.__slots__ + ('deep_copy',)
    type = NotImplemented  # There is no appropriate type, NotImplemented is handled specifically
    type_repr: str = 'Any'
    atomic = False

    def __init__(self, *, deep_copy: bool = False, default: Any = Missing, hide_none: bool = False,
                 primitive_name: Optional[str] = Missing, to_primitive_name: Optional[str] = Missing,
                 validators: Optional[List[Validator]] = None):
        super().__init__(default=default, hide_none=hide_none,
                         primitive_name=primitive_name, to_primitive_name=to_primitive_name, validators=validators)
        self.deep_copy = deep_copy

    def init_from_annotation(self, parser: AnnotationResolver):
        pass  # Don't require typing.Any, explicit AnyField should allow any typing

    def convert(self, value: Any) -> Any:
        if value is Missing:
            return self._fill_missing()
        return deepcopy(value) if self.deep_copy else value

    def copy_value(self, value: Any) -> Any:
        return deepcopy(value)

    def to_primitive(self, value: Any, role: Role = DEFAULT_ROLE, context: ToPrimitiveContextType = None) -> Any:
        return deepcopy(value) if self.deep_copy else value



from stereotype import Model, StrField, DEFAULT_ROLE


class Leaf(Model):
    type = 'leaf'

    color: str = 'green'


class PrivateStrField(StrField):
    def to_primitive(self, value, role=DEFAULT_ROLE, context=None):
        if context is not None and context.get('private', False):
            return '<hidden>'
        return value



from __future__ import annotations

from typing import Optional, Tuple, List, Iterable, Type, Set, Any, Callable

from stereotype.fields.base import Field
from stereotype.meta import ModelMeta
from stereotype.roles import Role, RequestedRoleFields, FinalizedRoleFields, DEFAULT_ROLE
from stereotype.utils import Missing, ValidationError, ConversionError, PathErrorType, ValidationContextType


class Model(metaclass=ModelMeta):
    """
    The base class for all your models.

    May inherit from other models, but if inheriting from multiple Models, all but one of the bases must be marked
    as abstract by adding the attribute ``__abstract__ = True``.

    Class attributes will become fields (deserialized, validated, serialized) if they:

    * are a public attribute (i.e. not prefixed with ``_``)
    * they have a type hint, unless that type hint is wrapped in ``ClassVar``
    """

    __slots__ = []
    # Don't use these fields in external code directly, they may not be initialized!
    __fields__: List[Field]
    __input_fields__: List[_InputFieldConfig]
    __validated_fields__: List[_ValidatedFieldConfig]
    __role_fields__: List[List[_OutputFieldConfig]]
    __roles__: List[FinalizedRoleFields]
    __gettable__: Set[str]

    def __init__(self, raw_data: Optional[dict] = None):
        """
        Constructs an instance of this Model from input data. Does not perform validation,
        but the conversion may raise a :class:`ConversionError` for a single field with a bad value.

        :param raw_data: Input data to convert, a dictionary with string keys and values corresponding to the fields.
        """
        if self.__input_fields__ is NotImplemented:
            self.__initialize_model__()
        if raw_data is None:
            raw_data = {}
        elif not isinstance(raw_data, dict):
            raise ConversionError.new(f'Supplied type {type(raw_data).__name__}, needs a mapping')

        for name, primitive_name, convert, copy_value in self.__input_fields__:
            if primitive_name is None:
                value = Missing
            else:
                value = raw_data.get(primitive_name, Missing)
            try:
                setattr(self, name, convert(value))
            except ConversionError as e:
                raise e.wrapped(primitive_name)
            except (TypeError, ValueError) as e:
                raise ConversionError.new(str(e), primitive_name)

    def to_primitive(self, role: Role = DEFAULT_ROLE, context=None):
        """
        Creates raw data from this Model, creating a copy of the data it was initialized with.

        :param role: Can be used together with declare_roles to exclude fields for certain roles.
        :param context: Optional value opaque to stereotype, useful for serializing custom field types.
        """
        if role.code < len(self.__role_fields__):
            fields = self.__role_fields__[role.code]
        elif role.empty_by_default:
            return {}
        else:
            fields = self.__role_fields__[0]

        result = {}
        for _, name, serializable, to_primitive, to_primitive_name, hide_none, hide_empty, empty_value in fields:
            if serializable is None:
                value = getattr(self, name)
                if value is Missing:
                    continue
                converted = to_primitive(value, role, context) if to_primitive is not None else value
                if (converted is None and hide_none) or (hide_empty and converted == empty_value):
                    continue
                result[to_primitive_name] = converted
            else:
                value = serializable(self)
                if value is None and hide_none:
                    continue
                result[to_primitive_name] = value
        return result

    serialize = to_primitive

    def validate(self, context=None):
        """
        Validates data, raising a :class:`ValidationError` with potentially multiple error messages,
        mapped by field paths.

        Validation catches: fields without defaults missing in data, None in non-Optional fields and failing validators
        (whether native Field validation options or custom validate_* methods).

        :param context: Optional value opaque to stereotype, passed to any custom validation methods (validate_*).
        """
        errors = list(self.validation_errors(context))
        if errors:
            raise ValidationError(errors)

    def validation_errors(self, context=None) -> Iterable[PathErrorType]:
        """
        Validates data, yielding errors one by one, each with a field path.

        Validation catches: fields without defaults missing in data, None in non-Optional fields and failing validators
        (whether native Field validation options, custom Field validators or custom validate_* methods).

        :param context: Optional value opaque to stereotype, passed to Field validators and validate_* Model methods.
        """
        for name, input_name, allow_none, validation_errors, validator_method in self.__validated_fields__:
            value = getattr(self, name)
            for path, error in validation_errors(value, context):
                yield (input_name,) + path, error
            if validator_method is not None and value is not Missing and (value is not None or allow_none):
                try:
                    validator_method(self, value, context)
                except ValueError as e:
                    yield (input_name,), str(e)

    @classmethod
    def declare_roles(cls) -> Iterable[RequestedRoleFields]:
        """
        Override to specify which fields are serialized for particular roles.

        Define a :class:`stereotype.Role` globally, like:  ``MY_ROLE = Role('my')``

        Then in this class method do for example:          ``yield MY_ROLE.blacklist(cls.my_field)``
        """
        yield from ()

    @classmethod
    def resolve_extra_types(cls) -> Set[Type[Model]]:
        """
        Override to provide a set of symbols that are not available globally in this model's module, such as
        locally declared other models or models that cannot be imported globally and use ``if TYPE_CHECKING``.
        Simply return any such model classes in this class method, importing them locally if needed.

        If inheriting from a Model that implements this, use ``return {...} | super().resolve_extra_types()``
        """
        return set()

    def __getitem__(self, key):
        if key not in self.__gettable__:
            raise KeyError(key)  # Avoids getting methods via getitem
        try:
            return getattr(self, key)
        except AttributeError:
            raise KeyError(key)  # Could happen for user-defined slot attributes that were never initialized

    def get(self, key, default=None):
        """
        Returns a value of a field with the given key, or the provided default if no such field exists
        or a required field is missing value.
        """
        if key not in self.__gettable__:
            return default
        value = getattr(self, key, Missing)
        return value if value is not Missing else default

    def __eq__(self, other: Model):
        if type(self) != type(other):
            return False
        for name, primitive_name, convert, copy_value in self.__input_fields__:
            if getattr(self, name) != getattr(other, name):
                return False
        return True

    def __repr__(self):
        parts = []
        for field in self.__fields__:
            base = f'{field.name}='
            if field.serializable is not None:
                continue
            value = getattr(self, field.name)
            if field.type is list or field.type is dict:
                if value:
                    if field.type is list:
                        parts.append(f'{base}[({len(value)} items)]')
                    else:
                        parts.append(f'{base}{{({len(value)} items)}}')
                else:
                    parts.append(f'{base}{value!r}')
            elif field.type is not NotImplemented and issubclass(field.type, Model):
                parts.append(f'{base}{type(value).__name__ if isinstance(value, Model) else value}')
            else:
                parts.append(f'{base}{repr(value)}')
        return f'<{self.__class__.__name__} {{' + ', '.join(parts) + '}>'

    def items(self) -> Iterable[Tuple[str, Any]]:
        """Provides an iterator over the fields of this model, with field name and converted value pairs."""
        for name, primitive_name, convert, copy_value in self.__input_fields__:
            value = getattr(self, name)
            if value is Missing:
                continue  # Missing required fields are skipped intentionally
            yield name, value

    def copy(self, deep: bool = False) -> Model:
        """Creates an optionally deep copy of this model."""
        copied = self.__new__(self.__class__)
        for name, primitive_name, convert, copy_value in self.__input_fields__:
            value = getattr(self, name)
            if deep and copy_value is not None:
                value = copy_value(value)
            setattr(copied, name, value)
        return copied

    @classmethod
    def fields_for_role(cls, role: Role = DEFAULT_ROLE) -> List[Field]:
        """
        Lists fields present in output for a given role. Omits those suppressed by roles or None `primitive_name`.

        :param role: The :class:`Role` that controls which fields are present
        """
        if cls.__role_fields__ is NotImplemented:
            cls.__initialize_model__()
        if role.code < len(cls.__role_fields__):
            return [field for field, _, _, _, _, _, _, _ in cls.__role_fields__[role.code]]
        else:
            return [] if role.empty_by_default else list(cls.__fields__)

    @classmethod
    def field_names_for_role(cls, role: Role = DEFAULT_ROLE) -> List[str]:
        """
        Lists the field names (using primitive names, as in serialized data) present for a given role.

        :param role: The :class:`Role` that controls which fields are present
        """
        return [field.to_primitive_name for field in cls.fields_for_role(role)]


_NativeValidator = Callable[[Any, ValidationContextType], Iterable[PathErrorType]]
_ValidatorMethod = Callable[[Model, Any, ValidationContextType], None]
_SerializableFn = Callable[[Model], Any]
_ToPrimitive = Callable[[Any, Role, Any], Any]

# These rather ugly tuples measurably improve performance compared to accessing field attributes
_InputFieldConfig = Tuple[
    str,  # name
    Optional[str],  # primitive_name - if None, the field may not be filled from input data
    Callable[[Any], Any],  # convert - turn data into the field's native representation
    Optional[Callable[[Any], Any]],  # copy_value - if None, the field is atomic and doesn't need copying
]
_ValidatedFieldConfig = Tuple[
    str,  # name
    str,  # input_name - used in error message, has fallback to name if primitive names are None
    bool,  # allow_none
    Optional[_NativeValidator],  # Field's validate method
    Optional[_ValidatorMethod],  # Model's validator_method bound to a Field
]
_OutputFieldConfig = Tuple[
    Field,
    str,  # name
    Optional[_SerializableFn],  # serializable - if provided, value of the field is extracted by calling this
    Optional[_ToPrimitive],  # to_primitive - if provided, should be called to convert native value to primitive
    Optional[str],  # to_primitive_name - key used for output; in practice, this tuple isn't used if None
    bool,  # hide_none
    bool,  # hide_empty - if True, don't add the key if the value equals empty_value
    Any,  # empty_value
]



from typing import Any, Optional
from unittest import TestCase

from stereotype import StrField, Missing, Role, DEFAULT_ROLE, Model


class PrefixStrField(StrField):
    """
    Converted data strips a prefix from primitive data.
    Demonstrates creating custom fields, especially with required parameters.
    """

    def __init__(self, prefix: str, **kwargs):
        super().__init__(**kwargs)
        self.prefix = prefix

    def convert(self, value: Any) -> Any:
        converted = super().convert(value)
        if value is None or value is Missing:
            return value
        return converted[len(self.prefix):] if converted.startswith(self.prefix) else value

    def to_primitive(self, value: Any, role: Role = DEFAULT_ROLE, context=None) -> Any:
        primitive = super().to_primitive(value, role, context)
        if value is None or value is Missing:
            return value
        return self.prefix + primitive


class TestCustomFields(TestCase):
    def test_prefix_str_field(self):
        class MyModel(Model):
            a_prefix: Optional[str] = PrefixStrField('a_')

        model = MyModel({"a_prefix": "a_value"})
        self.assertEqual("value", model.a_prefix)
        self.assertEqual({"a_prefix": "a_value"}, model.to_primitive())

        model = MyModel({"a_prefix": "value"})
        self.assertEqual("value", model.a_prefix)
        self.assertEqual({"a_prefix": "a_value"}, model.to_primitive())

        model = MyModel({"a_prefix": None})
        self.assertIsNone(model.a_prefix)
        self.assertEqual({"a_prefix": None}, model.to_primitive())

        model = MyModel({})
        self.assertIs(Missing, model.a_prefix)
        self.assertEqual({}, model.to_primitive())



from __future__ import annotations

import sys
from typing import Set, Type, Union
from unittest import TestCase, skip

from stereotype import Model, DictField, IntField, ValidationError, DynamicModelField

if tuple(sys.version_info[:2]) < (3, 10):
    skip_if_not_python3_10 = skip('This test suite only works in Python 3.10 and above')
else:
    def skip_if_not_python3_10(cls):
        return cls


@skip_if_not_python3_10
class TestPython310Annotations(TestCase):
    def test_optional(self):
        class Optional(Model):
            opt: int | None = 1

        model = Optional({'opt': None})
        model.validate()
        self.assertEqual({'opt': None}, model.to_primitive())

    def test_list_dict(self):
        class Collections(Model):
            list: list[int] = []
            dict: dict[str, int] = DictField(value_field=IntField(max_value=5))

        model = Collections({'list': ['1', 2.0], 'dict': {}})
        model.validate()
        self.assertEqual({'list': [1, 2], 'dict': {}}, model.to_primitive())

        model = Collections({'dict': {'1': 1, 3: '3', '6': 6}})
        with self.assertRaises(ValidationError) as ctx:
            model.validate()
        self.assertEqual({'dict': {'6': ['Must be at most 5']}}, ctx.exception.errors)
        self.assertEqual({'list': [], 'dict': {'1': 1, '3': 3, '6': 6}}, model.to_primitive())

    def test_union(self):
        class A(Model):
            type = 'a'
            a: int = 1

        class B(Model):
            type = 'b'
            b: int = 2

        class Unionized(Model):
            old: Union[A, B] = A
            simple: A | B
            advanced: A | B = DynamicModelField(default=A, primitive_name='other')

            @classmethod
            def resolve_extra_types(cls) -> Set[Type[Model]]:
                return {A, B}

        model = Unionized({'simple': {'type': 'b'}})
        model.validate()
        self.assertEqual({
            'old': {'type': 'a', 'a': 1},
            'simple': {'type': 'b', 'b': 2},
            'other': {'type': 'a', 'a': 1},
        }, model.to_primitive())



# Parts of this file are referenced from docs by line numbers, always update docs after changing it!

from typing import Optional, ClassVar
from stereotype import Model, FloatField, ValidationError, Missing, ConversionError


class Employee(Model):
    name: str
    department: str = 'Engineering'
    female: Optional[bool] = None
    salary: float = FloatField(min_value=42., default=42.)

    # These attributes won't become fields
    human = True
    alive: ClassVar[bool] = True

    def greeting(self) -> str:
        title = {True: "Ms. ", False: "Mr. ", None: ""}[self.female]
        return f'Dear {title}{self.name}'


model = Employee({'name': 'Jane Doe', 'female': True, 'salary': 150000})
model.validate()
assert model.name == 'Jane Doe'
assert model.department == 'Engineering'
assert model.female is True
assert model.salary == 150000.0
assert model.human is True
assert model.alive is True
assert model.greeting() == 'Dear Ms. Jane Doe'
model.salary = 200000.0
primitive = model.to_primitive()
assert primitive == {'name': 'Jane Doe', 'department': 'Engineering', 'female': True, 'salary': 200000.0}


model = Employee({'department': None, 'salary': '10'})
try:
    model.validate()
    assert False, 'Validation fails due to errors below'  # pragma: no cover
except ValidationError as e:
    assert str(e) == 'name: This field is required'
    assert e.errors == {
        'name': ['This field is required'],
        'department': ['This field is required'],
        'salary': ['Must be at least 42.0'],
    }
assert model.name is Missing
assert model.department is None
assert model.salary == 10.0
assert model.to_primitive() == {'department': None, 'female': None, 'salary': 10.0}


try:
    model = Employee({'name': 'Alien', 'female': 'Unknown'})
except ConversionError as e:
    assert str(e) == 'female: Value must be a boolean or a true/false/yes/no string value'



from __future__ import annotations

from copy import copy, deepcopy
from typing import Set, cast, Any, Optional, List, Dict, Union, Type, ClassVar
from unittest import TestCase

from stereotype import Model, Missing, ValidationError, ConversionError, BoolField, IntField, ConfigurationError, \
    FloatField, StrField, DataError, serializable
from stereotype.fields.base import Field, AnyField
from stereotype.fields.compound import DictField
from tests.common import Leaf


class DriedLeaf(Leaf):
    healthy = False
    crunchy: ClassVar[bool] = True
    age: float = FloatField(max_value=5, default=0.)


class PropertyAccessBase(Model):
    __slots__ = ['extra_slot']

    basic_field: str

    @serializable
    def basic_serializable(self):
        return self.basic_field

    @property
    def basic_property(self):
        return self.basic_field


class PropertyAccess(PropertyAccessBase):
    default_field: str = 'hello'

    @serializable
    def default_serializable(self):
        return self.default_field

    @property
    def default_property(self):
        return self.default_field


class NonModel:
    def __init__(self, x=1):
        self.x = x

    def __eq__(self, other):
        return self.x == other.x

    def __repr__(self):
        return f'<{self.x}>'


class TestModels(TestCase):
    def test_bad_init(self):
        with self.assertRaises(ConversionError) as ctx:
            DriedLeaf(cast(dict, 'bad'))
        self.assertEqual('Supplied type str, needs a mapping', str(ctx.exception))
        self.assertEqual({'_global': ['Supplied type str, needs a mapping']}, ctx.exception.errors)

    def test_initialized_baseclass(self):
        Leaf()  # Ensures it is pre-initialized
        model = DriedLeaf({'color': 'brown'})
        self.assertEqual('brown', model.color)
        self.assertEqual(0, model.age)
        self.assertIs(False, model.healthy)
        self.assertIs(True, model.crunchy)
        model.validate()
        self.assertEqual({'color': 'brown', 'age': 0}, model.serialize())

    def test_non_initialized_baseclass(self):
        class MyBase(Model):
            a: int
            b: float
            c: int = IntField(hide_none=True, default=-5)

        class MyChild(MyBase):
            b: str = StrField(primitive_name='bb')
            d: bool = True

            def validate_a(self, value, _):
                if value <= 0:
                    raise ValueError('Must be positive')

            def validate_b(self, value, _):
                if not value:
                    raise ValueError('Must be truthy')

            def validate_c(self, value, _):
                if value >= 0:
                    raise ValueError('Must be negative')

        class AnotherChild(MyBase):
            pass

        model = MyChild({'a': 1, 'b': 'ignored', 'bb': 'abc'})
        self.assertEqual(1, model.a)
        self.assertEqual('abc', model.b)
        self.assertEqual(-5, model.c)
        self.assertIs(True, model.d)
        model.validate()
        self.assertEqual({'a': 1, 'bb': 'abc', 'c': -5, 'd': True}, model.serialize())
        model.a = 0
        model.b = ''
        model.c = 4
        with self.assertRaises(ValidationError) as ctx:
            model.validate()
        self.assertEqual({'a': ['Must be positive'], 'bb': ['Must be truthy'], 'c': ['Must be negative']},
                         ctx.exception.errors)

        # If explicit fields were not copied, the validator method may leak to the base class or the other subclass
        base = MyBase({'a': 0, 'b': 0, 'c': 1})
        self.assertEqual(0, base.a)
        self.assertEqual(0, base.b)
        self.assertEqual(1, base.c)
        base.validate()
        self.assertEqual({'a': 0, 'b': 0, 'c': 1}, base.serialize())

        other = AnotherChild({'a': 0, 'b': 0, 'c': 1})
        other.validate()

    def test_bad_field_type_typing(self):
        class BadType(Model):
            set: Set[int]

        with self.assertRaises(ConfigurationError) as ctx:
            BadType()
        self.assertEqual("Field set of BadType: Unrecognized field annotation typing.Set[int] "
                         "(may need an explicit Field)", str(ctx.exception))

    def test_bad_field_type_native(self):
        class BadType(Model):
            bad: complex

        with self.assertRaises(ConfigurationError) as ctx:
            BadType()
        self.assertEqual("Field bad of BadType: Unrecognized field annotation complex (may need an explicit Field)",
                         str(ctx.exception))

    def test_multiple_non_abstract_bases(self):
        class Base1(Model):
            a: int

        class Base2(Model):
            b: int

        with self.assertRaises(ConfigurationError) as ctx:
            class Parent(Base1, Base2):
                unused: Parent
        self.assertEqual('Parent: multiple bases have instance lay-out conflict, if inheriting from multiple models, '
                         'only one may have __slots__ (declare abstract models without __slots__ by adding class '
                         'attribute `__abstract__ = True`)', str(ctx.exception))

    def test_none_primitive_name(self):
        class NonePrimitive(Model):
            normal: bool = BoolField(primitive_name='ok', to_primitive_name='ordinary')
            none_both: int = IntField(primitive_name=None)
            none_input: float = FloatField(primitive_name=None, to_primitive_name='no_input')
            none_output: str = StrField(to_primitive_name=None)

        model = NonePrimitive({'ok': True, 'none_both': 1, 'none_input': 4.2, 'no_input': 4.7, 'none_output': 'data'})
        self.assertIs(True, model.normal)
        self.assertIs(Missing, model.none_both)
        self.assertIs(Missing, model.none_input)
        self.assertEqual('data', model.none_output)
        self.assertEqual({'ordinary': True}, model.serialize())

        model.none_both = 5
        model.none_input = 42.47
        model.none_output = 'changed'
        self.assertEqual({'ordinary': True, 'no_input': 42.47}, model.serialize())
        self.assertEqual(5, model.none_both)
        self.assertEqual('changed', model.none_output)

    def test_equality(self):
        model = DriedLeaf({'age': 4.2})
        self.assertEqual(DriedLeaf({'color': 'green', 'age': 4.2}), model)
        self.assertNotEqual(DriedLeaf({'color': 'green', 'age': 4.7}), model)
        self.assertNotEqual(DriedLeaf({'color': None, 'age': 4.2}), model)
        self.assertNotEqual(Leaf({'color': 'green', 'age': 4.2}), model)
        self.assertNotEqual(1, model)
        self.assertNotEqual(None, model)

    def test_override_field_type(self):
        class First(Model):
            plain: int
            custom: int = IntField(min_value=1)

        class Second(First):
            plain: float
            custom: float = FloatField(min_value=2)

        self.assertEqual({'plain': 1.1, 'custom': 2.2}, Second({'plain': '1.1', 'custom': '2.2'}).serialize())

    def test_resolve_extra_types_error(self):
        class Inner(Model):
            pass

        class Outer(Model):
            field: Inner

        with self.assertRaises(ConfigurationError) as e:
            Outer()
        self.assertEqual("Model Outer annotation name 'Inner' is not defined. If not a global symbol or cannot be "
                         "imported globally, use the class method `resolve_extra_types` to provide it.",
                         str(e.exception))

    def test_copy(self):
        class Dummy(Model):
            type = 'dummy'
            field: List[int] = list

        class CopyRight(Model):
            atomic: float
            any: Any
            atomic_list: List[str]
            complex_list: Optional[List[List[bool]]] = None
            atomic_dict: Dict[int, bool] = DictField(hide_none=True)
            complex_dict: Dict[bool, Dummy]
            model: Dummy
            dynamic_model: Optional[Union[Dummy, Leaf]]
            missing: Any = AnyField(deep_copy=True)

            @classmethod
            def resolve_extra_types(cls) -> Set[Type[Model]]:
                return {Dummy}

        model = CopyRight({'any': None, 'atomic_dict': None, 'dynamic_model': None})
        copied = model.copy(deep=True)
        model.atomic = 4.2
        model.any = {1: {True: False}}
        model.atomic_list = ['x']
        model.complex_list = [[]]
        model.atomic_dict = {1: True}
        model.complex_dict = {True: Dummy({'field': [1]})}
        model.model = None
        model.dynamic_model = Dummy()
        self.assertEqual({'any': None, 'complex_list': None, 'dynamic_model': None}, copied.serialize())

        model.model = Dummy({'field': [1, 2]})
        deep_copy = model.copy(deep=True)
        shallow_copy = model.copy(deep=False)
        model.atomic = 4.7
        model.any[1][False] = True
        model.any = Missing
        model.atomic_list.pop(-1)
        model.complex_list[0].append(True)
        model.complex_list.extend([[False], [True, False]])
        model.atomic_dict.pop(1)
        model.complex_dict[True].field.append(2)
        model.model.field.clear()
        model.dynamic_model.field.append(3)
        self.assertEqual({
            'atomic': 4.7,
            'atomic_list': [],
            'complex_list': [[True], [False], [True, False]],
            'atomic_dict': {},
            'complex_dict': {True: {'field': [1, 2]}},
            'model': {'field': []},
            'dynamic_model': {'type': 'dummy', 'field': [3]},
        }, model.to_primitive())
        self.assertEqual({
            'atomic': 4.2,
            'any': {1: {True: False}},
            'atomic_list': ['x'],
            'complex_list': [[]],
            'atomic_dict': {1: True},
            'complex_dict': {True: {'field': [1]}},
            'model': {'field': [1, 2]},
            'dynamic_model': {'type': 'dummy', 'field': []},
        }, deep_copy.serialize())
        self.assertEqual({
            'atomic': 4.2,
            'any': {1: {True: False, False: True}},
            'atomic_list': [],
            'complex_list': [[True], [False], [True, False]],
            'atomic_dict': {},
            'complex_dict': {True: {'field': [1, 2]}},
            'model': {'field': []},
            'dynamic_model': {'type': 'dummy', 'field': [3]},
        }, shallow_copy.serialize())

        self.assertIs(Missing, shallow_copy.missing)
        self.assertIs(Missing, deep_copy.missing)
        self.assertIs(Missing, copy(shallow_copy.missing))

    def test_any_field(self):
        class WithAny(Model):
            normal: Any
            optional: Optional[Any] = None
            custom: Optional[Any] = AnyField(deep_copy=True, default='x', hide_none=True, primitive_name='customized')
            non_model: Optional[NonModel] = AnyField(default=NonModel, hide_none=True)

            def validate_optional(self, value, _):
                if value:
                    raise ValueError('Must be falsy')

        initial = {'normal': {'x': 1}, 'customized': [1, [], {}], 'non_model': NonModel(x=42)}
        model = WithAny(initial)
        self.assertEqual({'x': 1}, model.normal)
        self.assertIsNone(model.optional)
        self.assertEqual([1, [], {}], model.custom)
        self.assertEqual(42, model.non_model.x)
        initial['customized'][1].append(2)
        initial['customized'].pop(-1)
        model.validate()
        self.assertEqual({
            'normal': {'x': 1},
            'optional': None,
            'customized': [1, [], {}],
            'non_model': NonModel(x=42),
        }, model.serialize())
        self.assertEqual("<WithAny {normal={'x': 1}, optional=None, custom=[1, [], {}], non_model=<42>}>", repr(model))

        model.normal['x'] = 'no copy'
        model.non_model = None
        self.assertEqual({'normal': {'x': 'no copy'}, 'optional': None, 'customized': [1, [], {}]}, model.serialize())
        model.serialize()['customized'][1].append(3)
        self.assertEqual([1, [], {}], model.custom)
        self.assertEqual(
            "<WithAny {normal={'x': 'no copy'}, optional=None, custom=[1, [], {}], non_model=None}>",
            repr(model),
        )

        model = WithAny({'optional': 123})
        self.assertEqual({'optional': 123, 'customized': 'x', 'non_model': NonModel(x=1)}, model.serialize())
        self.assertEqual("<WithAny {normal=Missing, optional=123, custom='x', non_model=<1>}>", repr(model))
        with self.assertRaises(ValidationError) as e:
            model.validate()
        self.assertEqual({
            'normal': ['This field is required'],
            'optional': ['Must be falsy'],
        }, e.exception.errors)

    def test_any_field_configuration_error_none_default(self):
        class Bad(Model):
            bad: Any = None

        with self.assertRaises(ConfigurationError) as e:
            Bad()
        self.assertEqual("Field bad of Bad: Cannot use None as default on a non-Optional Field", str(e.exception))

    def test_inheritance_from_non_model(self):
        class NonModel:
            some_attribute: str = "abc"

        class BaseModel(Model):
            x: int

        class MyModel(BaseModel, NonModel):
            some_field: str = "xyz"

        model = MyModel({"some_attribute": "ignore me", "x": 42})
        self.assertEqual("abc", model.some_attribute)
        self.assertEqual(42, model.x)
        self.assertEqual("xyz", model.some_field)
        model.validate()
        self.assertEqual({"x": 42, "some_field": "xyz"}, model.serialize())

    def test_abstract_model_from_concrete_with_slots(self):
        class Concrete(Model):
            x: int

        class Abstract(Concrete):
            __abstract__ = True
            __slots__ = ['a']
            y: float

        class AnotherConcrete(Concrete):
            __slots__ = ['b']
            z: str

        class Merged(Abstract, AnotherConcrete):
            __slots__ = ['c']

        model = Merged({'x': 1.0, 'y': 2, 'z': 3, 'a': 'ignore'})
        self.assertEqual(1, model.x)
        self.assertEqual(2.0, model.y)
        self.assertEqual('3', model.z)
        model.a = model.b = model.c = 'have slots'
        self.assertEqual('have slots', model.a)
        self.assertEqual('have slots', model.b)
        self.assertEqual('have slots', model.c)
        self.assertEqual({'x': 1, 'y': 2.0, 'z': '3'}, model.to_primitive())

    def test_resolve_extra_types_inheritance(self):
        class A(Model):
            a: int = 1

        class B(Model):
            b: float = 2.

        class X(Model):
            a: A = A

            @classmethod
            def resolve_extra_types(cls) -> Set[Type[Model]]:
                return {A}

        class Y(X):
            b: B = B

            @classmethod
            def resolve_extra_types(cls) -> Set[Type[Model]]:
                return {B} | super().resolve_extra_types()

        self.assertEqual({'a': {'a': 1}, 'b': {'b': 2}}, Y().to_primitive())

    def test_missing_singleton_copy(self):
        class ProduceMissing(Model):
            required: int

        model = ProduceMissing()
        self.assertIs(Missing, model.required)
        self.assertIs(Missing, copy(model).required)
        self.assertIs(Missing, deepcopy(model).required)
        self.assertEqual({}, model.serialize())
        self.assertEqual({}, copy(model).serialize())
        self.assertEqual({}, deepcopy(model).serialize())

    def test_explicit_field_no_annotation_error(self):
        with self.assertRaises(ConfigurationError) as ctx:
            class NoAnnotation(Model):
                some_field = StrField(max_length=5)

            _ = NoAnnotation
        self.assertEqual("Field some_field of Model class NoAnnotation defines an explicit Field "
                         "but lacks a type annotation or isn't public", str(ctx.exception))
        with self.assertRaises(ConfigurationError) as ctx:
            class Private(Model):
                _private: str = StrField(max_length=5)

            _ = Private
        self.assertEqual("Field _private of Model class Private defines an explicit Field "
                         "but lacks a type annotation or isn't public", str(ctx.exception))

    def test_get_field(self):
        model = PropertyAccess({'basic_field': 'help'})
        model.extra_slot = 'slot'

        # With value
        self.assertEqual('help', model.get('basic_field'))
        self.assertEqual('help', model.get('basic_serializable'))
        self.assertEqual('help', model.get('basic_property'))
        self.assertEqual('slot', model.get('extra_slot'))

        # With default value
        self.assertEqual('hello', model.get('default_field'))
        self.assertEqual('hello', model.get('default_serializable'))
        self.assertEqual('hello', model.get('default_property'))

        # Missing value
        incomplete_model = PropertyAccess()
        self.assertIsNone(incomplete_model.get('basic_field'))
        self.assertEqual('unknown', incomplete_model.get('basic_field', 'unknown'))
        self.assertIsNone(incomplete_model.get('basic_serializable'))
        self.assertEqual('unknown', incomplete_model.get('basic_serializable', 'unknown'))
        self.assertIsNone(incomplete_model.get('basic_property'))
        self.assertEqual('unknown', incomplete_model.get('basic_property', 'unknown'))

        # Non-existent field, method or unused slot
        incomplete_model = PropertyAccess()
        self.assertIsNone(incomplete_model.get('doesnt_exist'))
        self.assertEqual('unknown', model.get('doesnt_exist', 'unknown'))
        self.assertIsNone(incomplete_model.get('get'))
        self.assertEqual('unknown', model.get('get', 'unknown'))
        self.assertIsNone(incomplete_model.get('extra_slot'))
        self.assertEqual('unknown', incomplete_model.get('extra_slot', 'unknown'))

    def test_getitem(self):
        model = PropertyAccess({'basic_field': 'help'})
        model.extra_slot = 'slot'

        # With value
        self.assertEqual('help', model['basic_field'])
        self.assertEqual('help', model['basic_serializable'])
        self.assertEqual('help', model['basic_property'])
        self.assertEqual('slot', model['extra_slot'])

        # With default value
        self.assertEqual('hello', model['default_field'])
        self.assertEqual('hello', model['default_serializable'])
        self.assertEqual('hello', model['default_property'])

        # Missing value
        incomplete_model = PropertyAccess()
        self.assertIs(Missing, incomplete_model['basic_field'])
        self.assertIs(Missing, incomplete_model['basic_serializable'])
        self.assertIs(Missing, incomplete_model['basic_property'])

        # Non-existent field, method or unused slot
        with self.assertRaisesRegex(KeyError, "'doesnt_exist'"):
            self.fail(f'should raise: {model["doesnt_exist"]}')
        with self.assertRaisesRegex(KeyError, "'get'"):
            self.fail(f'should raise: {model["get"]}')
        with self.assertRaisesRegex(KeyError, "'extra_slot'"):
            self.fail(f'should raise: {incomplete_model["extra_slot"]}')

    def test_ensure_missing_coverage(self):
        # The only purpose of this test is to ensure 100% coverage for *dead* code, where possible
        self.assertEqual(1, IntField().to_primitive(1))

        class Temp(Model):
            field: int

        Temp()

        # noinspection PyAbstractClass
        class FakeField(Field):
            atomic = False
            type = set

        fake_field = FakeField()
        fake_field.name = 'field'
        list(fake_field.validate(42, {}))
        fake_field.copy_value(42)
        Temp.__fields__[0] = fake_field

        with self.assertRaises(AssertionError):
            DataError([])
